package com.sun.org.apache.bcel.internal.classfile;

import java.io.Writer;
import java.io.FilterWriter;
import java.io.Reader;
import java.io.FilterReader;
import java.util.zip.GZIPInputStream;
import java.io.ByteArrayInputStream;
import java.io.CharArrayReader;
import java.io.CharArrayWriter;
import java.util.zip.GZIPOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.util.Locale;
import java.util.List;
import java.util.ArrayList;
import java.io.IOException;
import com.sun.org.apache.bcel.internal.util.ByteSequence;
import com.sun.org.apache.bcel.internal.Const;

public abstract class Utility
{
    private static ThreadLocal<Integer> consumed_chars;
    private static boolean wide;
    private static final int FREE_CHARS = 48;
    private static int[] CHAR_MAP;
    private static int[] MAP_CHAR;
    private static final char ESCAPE_CHAR = '$';
    
    public Utility() {
    }
    
    private static int unwrap(ThreadLocal<Integer> tl) {
        return (int)tl.get();
    }
    
    private static void wrap(ThreadLocal<Integer> tl, int value) {
        tl.set(Integer.valueOf(value));
    }
    
    public static String accessToString(int access_flags) {
        return accessToString(access_flags, false);
    }
    
    public static String accessToString(int access_flags, boolean for_class) {
        StringBuilder buf = new StringBuilder();
        for (int p = 0, i = 0; p < 16384; ++i) {
            p = pow2(i);
            if ((access_flags & p) != 0x0) {
                if (for_class) {
                    if (p != 32) {
                        if (p == 512) {
                            continue;
                        }
                    }
                    else {
                        continue;
                    }
                }
                buf.append(Const.getAccessName(i)).append(" ");
            }
        }
        return buf.toString().trim();
    }
    
    public static String classOrInterface(int access_flags) {
        return ((access_flags & 0x200) != 0x0) ? "interface" : "class";
    }
    
    public static String codeToString(byte[] code, ConstantPool constant_pool, int index, int length, boolean verbose) {
        StringBuilder buf = new StringBuilder(code.length * 20);
        try (final ByteSequence stream = new ByteSequence(code)) {
            for (int i = 0; i < index; ++i) {
                codeToString(stream, constant_pool, verbose);
            }
            for (int i = 0; stream.available() > 0; ++i) {
                if (length < 0 || i < length) {
                    String indices = fillup(stream.getIndex() + ":", 6, true, ' ');
                    buf.append(indices).append(codeToString(stream, constant_pool, verbose)).append('\n');
                }
            }
        }
        catch (IOException e) {
            throw new ClassFormatException("Byte code error: " + buf.toString(), e);
        }
        return buf.toString();
    }
    
    public static String codeToString(byte[] code, ConstantPool constant_pool, int index, int length) {
        return codeToString(code, constant_pool, index, length, true);
    }
    
    public static String codeToString(ByteSequence bytes, ConstantPool constant_pool, boolean verbose) throws IOException {
        short opcode = (short)bytes.readUnsignedByte();
        int default_offset = 0;
        int no_pad_bytes = 0;
        StringBuilder buf = new StringBuilder(Const.getOpcodeName((int)opcode));
        if (opcode == 170 || opcode == 171) {
            int remainder = bytes.getIndex() % 4;
            no_pad_bytes = ((remainder == 0) ? 0 : (4 - remainder));
            for (int i = 0; i < no_pad_bytes; ++i) {
                byte b;
                if ((b = bytes.readByte()) != 0) {
                    System.err.println("Warning: Padding byte != 0 in " + Const.getOpcodeName((int)opcode) + ":" + (int)b);
                }
            }
            default_offset = bytes.readInt();
        }
        switch (opcode) {
            case 170: {
                int low = bytes.readInt();
                int high = bytes.readInt();
                int offset = bytes.getIndex() - 12 - no_pad_bytes - 1;
                default_offset += offset;
                buf.append("\tdefault = ").append(default_offset).append(", low = ").append(low).append(", high = ").append(high).append("(");
                int[] jump_table = new int[high - low + 1];
                for (int j = 0; j < jump_table.length; ++j) {
                    buf.append(jump_table[j] = offset + bytes.readInt());
                    if (j < jump_table.length - 1) {
                        buf.append(", ");
                    }
                }
                buf.append(")");
                break;
            }
            case 171: {
                int npairs = bytes.readInt();
                int offset = bytes.getIndex() - 8 - no_pad_bytes - 1;
                int[] match = new int[npairs];
                int[] jump_table = new int[npairs];
                default_offset += offset;
                buf.append("\tdefault = ").append(default_offset).append(", npairs = ").append(npairs).append(" (");
                for (int j = 0; j < npairs; ++j) {
                    match[j] = bytes.readInt();
                    jump_table[j] = offset + bytes.readInt();
                    buf.append("(").append(match[j]).append(", ").append(jump_table[j]).append(")");
                    if (j < npairs - 1) {
                        buf.append(", ");
                    }
                }
                buf.append(")");
                break;
            }
            case 153:
            case 154:
            case 155:
            case 156:
            case 157:
            case 158:
            case 159:
            case 160:
            case 161:
            case 162:
            case 163:
            case 164:
            case 165:
            case 166:
            case 167:
            case 168:
            case 198:
            case 199: {
                buf.append("\t\t#").append(bytes.getIndex() - 1 + bytes.readShort());
                break;
            }
            case 200:
            case 201: {
                buf.append("\t\t#").append(bytes.getIndex() - 1 + bytes.readInt());
                break;
            }
            case 21:
            case 22:
            case 23:
            case 24:
            case 25:
            case 54:
            case 55:
            case 56:
            case 57:
            case 58:
            case 169: {
                int vindex;
                if (Utility.wide) {
                    vindex = bytes.readUnsignedShort();
                    Utility.wide = false;
                }
                else {
                    vindex = bytes.readUnsignedByte();
                }
                buf.append("\t\t%").append(vindex);
                break;
            }
            case 196: {
                Utility.wide = true;
                buf.append("\t(wide)");
                break;
            }
            case 188: {
                buf.append("\t\t<").append(Const.getTypeName((int)bytes.readByte())).append(">");
                break;
            }
            case 178:
            case 179:
            case 180:
            case 181: {
                int index = bytes.readUnsignedShort();
                buf.append("\t\t").append(constant_pool.constantToString(index, (byte)9)).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 187:
            case 192: {
                buf.append("\t");
            }
            case 193: {
                int index = bytes.readUnsignedShort();
                buf.append("\t<").append(constant_pool.constantToString(index, (byte)7)).append(">").append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 183:
            case 184: {
                int index = bytes.readUnsignedShort();
                Constant c = constant_pool.getConstant(index);
                buf.append("\t").append(constant_pool.constantToString(index, c.getTag())).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 182: {
                int index = bytes.readUnsignedShort();
                buf.append("\t").append(constant_pool.constantToString(index, (byte)10)).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 185: {
                int index = bytes.readUnsignedShort();
                int nargs = bytes.readUnsignedByte();
                buf.append("\t").append(constant_pool.constantToString(index, (byte)11)).append(verbose ? (" (" + index + ")\t") : "").append(nargs).append("\t").append(bytes.readUnsignedByte());
                break;
            }
            case 186: {
                int index = bytes.readUnsignedShort();
                buf.append("\t").append(constant_pool.constantToString(index, (byte)18)).append(verbose ? (" (" + index + ")\t") : "").append(bytes.readUnsignedByte()).append(bytes.readUnsignedByte());
                break;
            }
            case 19:
            case 20: {
                int index = bytes.readUnsignedShort();
                buf.append("\t\t").append(constant_pool.constantToString(index, constant_pool.getConstant(index).getTag())).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 18: {
                int index = bytes.readUnsignedByte();
                buf.append("\t\t").append(constant_pool.constantToString(index, constant_pool.getConstant(index).getTag())).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 189: {
                int index = bytes.readUnsignedShort();
                buf.append("\t\t<").append(compactClassName(constant_pool.getConstantString(index, (byte)7), false)).append(">").append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 197: {
                int index = bytes.readUnsignedShort();
                int dimensions = bytes.readUnsignedByte();
                buf.append("\t<").append(compactClassName(constant_pool.getConstantString(index, (byte)7), false)).append(">\t").append(dimensions).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 132: {
                int vindex;
                int constant;
                if (Utility.wide) {
                    vindex = bytes.readUnsignedShort();
                    constant = bytes.readShort();
                    Utility.wide = false;
                }
                else {
                    vindex = bytes.readUnsignedByte();
                    constant = bytes.readByte();
                }
                buf.append("\t\t%").append(vindex).append("\t").append(constant);
                break;
            }
            default: {
                if (Const.getNoOfOperands((int)opcode) > 0) {
                    for (int k = 0; (long)k < Const.getOperandTypeCount((int)opcode); ++k) {
                        buf.append("\t\t");
                        switch (Const.getOperandType((int)opcode, k)) {
                            case 8: {
                                buf.append((int)bytes.readByte());
                                break;
                            }
                            case 9: {
                                buf.append((int)bytes.readShort());
                                break;
                            }
                            case 10: {
                                buf.append(bytes.readInt());
                                break;
                            }
                            default: {
                                throw new IllegalStateException("Unreachable default case reached!");
                            }
                        }
                    }
                    break;
                }
                break;
            }
        }
        return buf.toString();
    }
    
    public static String codeToString(ByteSequence bytes, ConstantPool constant_pool) throws IOException {
        return codeToString(bytes, constant_pool, true);
    }
    
    public static String compactClassName(String str) {
        return compactClassName(str, true);
    }
    
    public static String compactClassName(String str, String prefix, boolean chopit) {
        int len = prefix.length();
        str = str.replace('/', '.');
        if (chopit && str.startsWith(prefix) && str.substring(len).indexOf(46) == -1) {
            str = str.substring(len);
            return str;
        }
        return str;
    }
    
    public static String compactClassName(String str, boolean chopit) {
        return compactClassName(str, "java.lang.", chopit);
    }
    
    public static int setBit(int flag, int i) {
        return flag | pow2(i);
    }
    
    public static int clearBit(int flag, int i) {
        int bit = pow2(i);
        return ((flag & bit) == 0x0) ? flag : (flag ^ bit);
    }
    
    public static boolean isSet(int flag, int i) {
        return (flag & pow2(i)) != 0x0;
    }
    
    public static String methodTypeToSignature(String ret, String[] argv) throws ClassFormatException {
        StringBuilder buf = new StringBuilder("(");
        if (argv != null) {
            int length = argv.length;
            int i = 0;
            while (i < length) {
                String element = argv[i];
                String str = getSignature(element);
                if (str.endsWith("V")) {
                    throw new ClassFormatException("Invalid type: " + element);
                }
                buf.append(str);
                ++i;
            }
        }
        String str = getSignature(ret);
        buf.append(")").append(str);
        return buf.toString();
    }
    
    public static String[] methodSignatureArgumentTypes(String signature) throws ClassFormatException {
        return methodSignatureArgumentTypes(signature, true);
    }
    
    public static String[] methodSignatureArgumentTypes(String signature, boolean chopit) throws ClassFormatException {
        List<String> vec = new ArrayList();
        try {
            if (signature.charAt(0) != '(') {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            for (int index = 1; signature.charAt(index) != ')'; index += unwrap(Utility.consumed_chars)) {
                vec.add(signatureToString(signature.substring(index), chopit));
            }
        }
        catch (StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
        return (String[])vec.toArray(new String[vec.size()]);
    }
    
    public static String methodSignatureReturnType(String signature) throws ClassFormatException {
        return methodSignatureReturnType(signature, true);
    }
    
    public static String methodSignatureReturnType(String signature, boolean chopit) throws ClassFormatException {
        String type;
        try {
            int index = signature.lastIndexOf(41) + 1;
            type = signatureToString(signature.substring(index), chopit);
        }
        catch (StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
        return type;
    }
    
    public static String methodSignatureToString(String signature, String name, String access) {
        return methodSignatureToString(signature, name, access, true);
    }
    
    public static String methodSignatureToString(String signature, String name, String access, boolean chopit) {
        return methodSignatureToString(signature, name, access, chopit, null);
    }
    
    public static String methodSignatureToString(String signature, String name, String access, boolean chopit, LocalVariableTable vars) throws ClassFormatException {
        StringBuilder buf = new StringBuilder("(");
        int var_index = access.contains("static") ? 0 : 1;
        String type;
        try {
            if (signature.charAt(0) != '(') {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            int index;
            for (index = 1; signature.charAt(index) != ')'; index += unwrap(Utility.consumed_chars)) {
                String param_type = signatureToString(signature.substring(index), chopit);
                buf.append(param_type);
                if (vars != null) {
                    LocalVariable l = vars.getLocalVariable(var_index, 0);
                    if (l != null) {
                        buf.append(" ").append(l.getName());
                    }
                }
                else {
                    buf.append(" arg").append(var_index);
                }
                if ("double".equals(param_type) || "long".equals(param_type)) {
                    var_index += 2;
                }
                else {
                    ++var_index;
                }
                buf.append(", ");
            }
            ++index;
            type = signatureToString(signature.substring(index), chopit);
        }
        catch (StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
        if (buf.length() > 1) {
            buf.setLength(buf.length() - 2);
        }
        buf.append(")");
        return access + ((access.length() > 0) ? " " : "") + type + " " + name + buf.toString();
    }
    
    private static int pow2(int n) {
        return 1 << n;
    }
    
    public static String replace(String str, String old, String new_) {
        try {
            if (str.contains(old)) {
                StringBuilder buf = new StringBuilder();
                int old_index;
                int index;
                for (old_index = 0; (index = str.indexOf(old, old_index)) != -1; old_index = index + old.length()) {
                    buf.append(str.substring(old_index, index));
                    buf.append(new_);
                }
                buf.append(str.substring(old_index));
                str = buf.toString();
            }
        }
        catch (StringIndexOutOfBoundsException e) {
            System.err.println(e);
        }
        return str;
    }
    
    public static String signatureToString(String signature) {
        return signatureToString(signature, true);
    }
    
    public static String signatureToString(String signature, boolean chopit) {
        wrap(Utility.consumed_chars, 1);
        try {
            switch (signature.charAt(0)) {
                case 'B': {
                    return "byte";
                }
                case 'C': {
                    return "char";
                }
                case 'D': {
                    return "double";
                }
                case 'F': {
                    return "float";
                }
                case 'I': {
                    return "int";
                }
                case 'J': {
                    return "long";
                }
                case 'T': {
                    int index = signature.indexOf(59);
                    if (index < 0) {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    wrap(Utility.consumed_chars, index + 1);
                    return compactClassName(signature.substring(1, index), chopit);
                }
                case 'L': {
                    int fromIndex = signature.indexOf(60);
                    if (fromIndex < 0) {
                        fromIndex = 0;
                    }
                    else {
                        fromIndex = signature.indexOf(62, fromIndex);
                        if (fromIndex < 0) {
                            throw new ClassFormatException("Invalid signature: " + signature);
                        }
                    }
                    int index2 = signature.indexOf(59, fromIndex);
                    if (index2 < 0) {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    int bracketIndex = signature.substring(0, index2).indexOf(60);
                    if (bracketIndex < 0) {
                        wrap(Utility.consumed_chars, index2 + 1);
                        return compactClassName(signature.substring(1, index2), chopit);
                    }
                    fromIndex = signature.indexOf(59);
                    if (fromIndex < 0) {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    if (fromIndex < bracketIndex) {
                        wrap(Utility.consumed_chars, fromIndex + 1);
                        return compactClassName(signature.substring(1, fromIndex), chopit);
                    }
                    StringBuilder type = new StringBuilder(compactClassName(signature.substring(1, bracketIndex), chopit)).append("<");
                    int consumed_chars = bracketIndex + 1;
                    if (signature.charAt(consumed_chars) == '+') {
                        type.append("? extends ");
                        ++consumed_chars;
                    }
                    else if (signature.charAt(consumed_chars) == '-') {
                        type.append("? super ");
                        ++consumed_chars;
                    }
                    if (signature.charAt(consumed_chars) == '*') {
                        type.append("?");
                        ++consumed_chars;
                    }
                    else {
                        type.append(signatureToString(signature.substring(consumed_chars), chopit));
                        consumed_chars += unwrap(Utility.consumed_chars);
                        wrap(Utility.consumed_chars, consumed_chars);
                    }
                    while (signature.charAt(consumed_chars) != '>') {
                        type.append(", ");
                        if (signature.charAt(consumed_chars) == '+') {
                            type.append("? extends ");
                            ++consumed_chars;
                        }
                        else if (signature.charAt(consumed_chars) == '-') {
                            type.append("? super ");
                            ++consumed_chars;
                        }
                        if (signature.charAt(consumed_chars) == '*') {
                            type.append("?");
                            ++consumed_chars;
                        }
                        else {
                            type.append(signatureToString(signature.substring(consumed_chars), chopit));
                            consumed_chars += unwrap(Utility.consumed_chars);
                            wrap(Utility.consumed_chars, consumed_chars);
                        }
                    }
                    ++consumed_chars;
                    type.append(">");
                    if (signature.charAt(consumed_chars) == '.') {
                        type.append(".");
                        type.append(signatureToString("L" + signature.substring(consumed_chars + 1), chopit));
                        consumed_chars += unwrap(Utility.consumed_chars);
                        wrap(Utility.consumed_chars, consumed_chars);
                        return type.toString();
                    }
                    if (signature.charAt(consumed_chars) != ';') {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    wrap(Utility.consumed_chars, consumed_chars + 1);
                    return type.toString();
                }
                case 'S': {
                    return "short";
                }
                case 'Z': {
                    return "boolean";
                }
                case '[': {
                    StringBuilder brackets = new StringBuilder();
                    int n;
                    for (n = 0; signature.charAt(n) == '['; ++n) {
                        brackets.append("[]");
                    }
                    int consumed_chars2 = n;
                    String type2 = signatureToString(signature.substring(n), chopit);
                    int _temp = unwrap(Utility.consumed_chars) + consumed_chars2;
                    wrap(Utility.consumed_chars, _temp);
                    return type2 + brackets.toString();
                }
                case 'V': {
                    return "void";
                }
                default: {
                    throw new ClassFormatException("Invalid signature: `" + signature + "'");
                }
            }
        }
        catch (StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid signature: " + signature, e);
        }
    }
    
    public static String getSignature(String type) {
        StringBuilder buf = new StringBuilder();
        char[] chars = type.toCharArray();
        boolean char_found = false;
        boolean delim = false;
        int index = -1;
    Label_0162:
        for (int i = 0; i < chars.length; ++i) {
            switch (chars[i]) {
                case '\t':
                case '\n':
                case '\f':
                case '\r':
                case ' ': {
                    if (char_found) {
                        delim = true;
                        break;
                    }
                    break;
                }
                case '[': {
                    if (!char_found) {
                        throw new RuntimeException("Illegal type: " + type);
                    }
                    index = i;
                    break Label_0162;
                }
                default: {
                    char_found = true;
                    if (!delim) {
                        buf.append(chars[i]);
                        break;
                    }
                    break;
                }
            }
        }
        int brackets = 0;
        if (index > 0) {
            brackets = countBrackets(type.substring(index));
        }
        type = buf.toString();
        buf.setLength(0);
        for (int j = 0; j < brackets; ++j) {
            buf.append('[');
        }
        boolean found = false;
        for (int k = 4; k <= 12 && !found; ++k) {
            if (Const.getTypeName(k).equals(type)) {
                found = true;
                buf.append(Const.getShortTypeName(k));
            }
        }
        if (!found) {
            buf.append('L').append(type.replace('.', '/')).append(';');
        }
        return buf.toString();
    }
    
    private static int countBrackets(String brackets) {
        char[] chars = brackets.toCharArray();
        int count = 0;
        boolean open = false;
        for (char c : chars) {
            switch (c) {
                case '[': {
                    if (open) {
                        throw new RuntimeException("Illegally nested brackets:" + brackets);
                    }
                    open = true;
                    break;
                }
                case ']': {
                    if (!open) {
                        throw new RuntimeException("Illegally nested brackets:" + brackets);
                    }
                    open = false;
                    ++count;
                    break;
                }
            }
        }
        if (open) {
            throw new RuntimeException("Illegally nested brackets:" + brackets);
        }
        return count;
    }
    
    public static byte typeOfMethodSignature(String signature) throws ClassFormatException {
        try {
            if (signature.charAt(0) != '(') {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            int index = signature.lastIndexOf(41) + 1;
            return typeOfSignature(signature.substring(index));
        }
        catch (StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
    }
    
    public static byte typeOfSignature(String signature) throws ClassFormatException {
        try {
            switch (signature.charAt(0)) {
                case 'B': {
                    return 8;
                }
                case 'C': {
                    return 5;
                }
                case 'D': {
                    return 7;
                }
                case 'F': {
                    return 6;
                }
                case 'I': {
                    return 10;
                }
                case 'J': {
                    return 11;
                }
                case 'L':
                case 'T': {
                    return 14;
                }
                case '[': {
                    return 13;
                }
                case 'V': {
                    return 12;
                }
                case 'Z': {
                    return 4;
                }
                case 'S': {
                    return 9;
                }
                case '!':
                case '*':
                case '+': {
                    return typeOfSignature(signature.substring(1));
                }
                default: {
                    throw new ClassFormatException("Invalid method signature: " + signature);
                }
            }
        }
        catch (StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
    }
    
    public static short searchOpcode(String name) {
        name = name.toLowerCase(Locale.ENGLISH);
        short i = 0;
        while (i < Const.OPCODE_NAMES_LENGTH) {
            if (Const.getOpcodeName((int)i).equals(name)) {
                return i;
            }
            ++i;
        }
        return -1;
    }
    
    private static short byteToShort(byte b) {
        return (b < 0) ? ((short)(256 + b)) : ((short)b);
    }
    
    public static String toHexString(byte[] bytes) {
        StringBuilder buf = new StringBuilder();
        for (int i = 0; i < bytes.length; ++i) {
            short b = byteToShort(bytes[i]);
            String hex = Integer.toHexString((int)b);
            if (b < 16) {
                buf.append('0');
            }
            buf.append(hex);
            if (i < bytes.length - 1) {
                buf.append(' ');
            }
        }
        return buf.toString();
    }
    
    public static String format(int i, int length, boolean left_justify, char fill) {
        return fillup(Integer.toString(i), length, left_justify, fill);
    }
    
    public static String fillup(String str, int length, boolean left_justify, char fill) {
        int len = length - str.length();
        char[] buf = new char[(len < 0) ? 0 : len];
        for (int j = 0; j < buf.length; ++j) {
            buf[j] = fill;
        }
        if (left_justify) {
            return str + new String(buf);
        }
        return new String(buf) + str;
    }
    
    static boolean equals(byte[] a, byte[] b) {
        int size;
        if ((size = a.length) != b.length) {
            return false;
        }
        int i = 0;
        while (i < size) {
            if (a[i] != b[i]) {
                return false;
            }
            ++i;
        }
        return true;
    }
    
    public static void printArray(PrintStream out, Object[] obj) {
        out.println(printArray(obj, true));
    }
    
    public static void printArray(PrintWriter out, Object[] obj) {
        out.println(printArray(obj, true));
    }
    
    public static String printArray(Object[] obj) {
        return printArray(obj, true);
    }
    
    public static String printArray(Object[] obj, boolean braces) {
        return printArray(obj, braces, false);
    }
    
    public static String printArray(Object[] obj, boolean braces, boolean quote) {
        if (obj == null) {
            return null;
        }
        StringBuilder buf = new StringBuilder();
        if (braces) {
            buf.append('{');
        }
        for (int i = 0; i < obj.length; ++i) {
            if (obj[i] != null) {
                buf.append(quote ? "\"" : "").append(obj[i]).append(quote ? "\"" : "");
            }
            else {
                buf.append("null");
            }
            if (i < obj.length - 1) {
                buf.append(", ");
            }
        }
        if (braces) {
            buf.append('}');
        }
        return buf.toString();
    }
    
    public static boolean isJavaIdentifierPart(char ch) {
        return (ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9') || ch == '_';
    }
    
    public static String encode(byte[] bytes, boolean compress) throws IOException {
        if (compress) {
            try (final ByteArrayOutputStream baos = new ByteArrayOutputStream();
                 final GZIPOutputStream gos = new GZIPOutputStream(baos)) {
                gos.write(bytes, 0, bytes.length);
                bytes = baos.toByteArray();
            }
        }
        CharArrayWriter caw = new CharArrayWriter();
        try (final Utility.JavaWriter jw = new Utility.JavaWriter(caw)) {
            for (byte b : bytes) {
                int in = b & 0xFF;
                jw.write(in);
            }
        }
        return caw.toString();
    }
    
    public static byte[] decode(String s, boolean uncompress) throws IOException {
        byte[] bytes;
        try (final Utility.JavaReader jr = new Utility.JavaReader(new CharArrayReader(s.toCharArray()));
             final ByteArrayOutputStream bos = new ByteArrayOutputStream()) {
            int ch;
            while ((ch = jr.read()) >= 0) {
                bos.write(ch);
            }
            bytes = bos.toByteArray();
        }
        if (uncompress) {
            GZIPInputStream gis = new GZIPInputStream(new ByteArrayInputStream(bytes));
            byte[] tmp = new byte[bytes.length * 3];
            int count = 0;
            int b;
            while ((b = gis.read()) >= 0) {
                tmp[count++] = (byte)b;
            }
            bytes = new byte[count];
            System.arraycopy(tmp, 0, bytes, 0, count);
            return bytes;
        }
        return bytes;
    }
    
    public static String convertString(String label) {
        char[] ch = label.toCharArray();
        StringBuilder buf = new StringBuilder();
        for (char element : ch) {
            switch (element) {
                case '\n': {
                    buf.append("\\n");
                    break;
                }
                case '\r': {
                    buf.append("\\r");
                    break;
                }
                case '\"': {
                    buf.append("\\\"");
                    break;
                }
                case '\'': {
                    buf.append("\\'");
                    break;
                }
                case '\\': {
                    buf.append("\\\\");
                    break;
                }
                default: {
                    buf.append(element);
                    break;
                }
            }
        }
        return buf.toString();
    }
    
    static {
        Utility.consumed_chars = new ThreadLocal<Integer>() {
            @Override
            protected Integer initialValue() {
                return Integer.valueOf(0);
            }
        };
        Utility.wide = false;
        Utility.CHAR_MAP = new int[48];
        Utility.MAP_CHAR = new int[256];
        int j = 0;
        for (int i = 65; i <= 90; ++i) {
            Utility.CHAR_MAP[j] = i;
            Utility.MAP_CHAR[i] = j;
            ++j;
        }
        for (int i = 103; i <= 122; ++i) {
            Utility.CHAR_MAP[j] = i;
            Utility.MAP_CHAR[i] = j;
            ++j;
        }
        Utility.CHAR_MAP[j] = 36;
        Utility.MAP_CHAR[36] = j;
        ++j;
        Utility.CHAR_MAP[j] = 95;
        Utility.MAP_CHAR[95] = j;
    }
    
    private static class JavaReader extends FilterReader
    {
        public JavaReader(Reader in) {
            super(in);
        }
        
        @Override
        public int read() throws IOException {
            int b = this.in.read();
            if (b != 36) {
                return b;
            }
            int i = this.in.read();
            if (i < 0) {
                return -1;
            }
            if ((i >= 48 && i <= 57) || (i >= 97 && i <= 102)) {
                int j = this.in.read();
                if (j < 0) {
                    return -1;
                }
                char[] tmp = { (char)i, (char)j };
                int s = Integer.parseInt(new String(tmp), 16);
                return s;
            }
            else {
                return Utility.MAP_CHAR[i];
            }
        }
        
        @Override
        public int read(char[] cbuf, int off, int len) throws IOException {
            for (int i = 0; i < len; ++i) {
                cbuf[off + i] = (char)this.read();
            }
            return len;
        }
    }
    
    private static class JavaWriter extends FilterWriter
    {
        public JavaWriter(Writer out) {
            super(out);
        }
        
        @Override
        public void write(int b) throws IOException {
            if (Utility.isJavaIdentifierPart((char)b) && b != 36) {
                this.out.write(b);
            }
            else {
                this.out.write(36);
                if (b >= 0 && b < 48) {
                    this.out.write(Utility.CHAR_MAP[b]);
                }
                else {
                    char[] tmp = Integer.toHexString(b).toCharArray();
                    if (tmp.length == 1) {
                        this.out.write(48);
                        this.out.write((int)tmp[0]);
                    }
                    else {
                        this.out.write((int)tmp[0]);
                        this.out.write((int)tmp[1]);
                    }
                }
            }
        }
        
        @Override
        public void write(char[] cbuf, int off, int len) throws IOException {
            for (int i = 0; i < len; ++i) {
                this.write((int)cbuf[off + i]);
            }
        }
        
        @Override
        public void write(String str, int off, int len) throws IOException {
            this.write(str.toCharArray(), off, len);
        }
    }
}
